
// screen is 2.4", 320x240

/*
 * comport messages
 * 
 * new - patch loaded
 * nxt - next/prev patch
 * nam - track name
 * key - transpose (-6 to 6)
 * vol - volume (0 - 1), controlled by slider
 * voo - actual volume (0 - 1)
 * vmn - min volume (0 - 1), used for automation
 * mod - mode (0 auto, 1 filter, 2 distortion, 3 min volume)
 * aon - automation on/off (1 / 0)
 * aam - automation speed (10 - 2000)
 * fon - filter on/off (1 / 0)
 * fam - filter amount (8 - 12500)
 * don - distortion on/off (1/0)
 * dam - distortion amount (received value 1 - 127, actual value 1-60)
 * 
 */

#include "SPI.h"
#include "ILI9341_t3.h"
//#include <CmdMessenger.h>  // CmdMessenger

// For the Adafruit shield, these are the default.
#define TFT_DC  9
#define TFT_CS 10

const int BCKGND_COLOUR = ILI9341_DARKGREY;
const int FOREGND_COLOUR = ILI9341_YELLOW;
const int ACTIVE_COLOUR = ILI9341_WHITE;
const int REC_COLOUR = ILI9341_RED;

const int LCD_FRAME_RATE = 30;
const int NO_OF_TRACKS = 8;
long previousMillis = 0;
String keys[12] = { "C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "Bb", "B" };

bool serialStarted = false;
bool serialEnded = false;
char inData[80];
int serIndex;

String current_patch_name = "";
String next_patch_name = "";
int current_key = -1;
float track_volumes[8] = { 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8 };
float track_auto_volumes[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
float track_min_volumes[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
String track_names[8] = { "", "", "", "", "", "", "", "" };
float track_auto_on[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
float track_auto_amount[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
float track_filter_on[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
float track_filter_amount[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
float track_distortion_on[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
float track_distortion_amount[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
float track_modes[8] = { 3, 3, 3, 3, 3, 3, 3, 3 };

String prev_current_patch_name = "";
String prev_next_patch_name = "";
int prev_current_key = -1;
float prev_track_volumes[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
float prev_track_auto_volumes[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
float prev_track_min_volumes[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
String prev_track_names[8] = { "", "", "", "", "", "", "", "" };
float prev_track_auto_on[8] = { -1, -1, -1, -1, -1, -1, -1, -1 };
float prev_track_auto_amount[8] = { 0, 0, 0, 0, 0, 1, 0, 0 };
float prev_track_filter_on[8] = { -1, -1, -1, -1, -1, -1, -1, -1 };
float prev_track_filter_amount[8] = { 0, 0, 0, 0, 0, 1, 0, 0 };
float prev_track_distortion_on[8] = { -1, -1, -1, -1, -1, -1, -1, -1 };
float prev_track_distortion_amount[8] = { 0, 0, 0, 0, 0, 1, 0, 0 };
float prev_track_modes[8] = { -1, -1, -1, -1, -1, -1, -1, -1 };

// Use hardware SPI (on Uno, #13, #12, #11) and the above for CS/DC
ILI9341_t3 tft = ILI9341_t3(TFT_CS, TFT_DC);
//Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS, TFT_DC, TFT_RST);

//CmdMessenger cmdMessenger = CmdMessenger(Serial);

void setup() {
  
  tft.begin();
  tft.setRotation(3);
  tft.fillScreen(ILI9341_BLACK);
  tft.setTextColor(ILI9341_YELLOW, ILI9341_BLACK);
  tft.setTextSize(2);
  tft.setCursor(10, 220); 
  tft.print("loading droninator...");

  Serial.begin(9600);
  while (!Serial) {
    ; // wait for serial port to connect. Needed for native USB port only
  }
  //while (!Serial) ; // wait for Arduino Serial Monitor
  Serial.println("Ready!"); // this seems to be required

  tft.fillScreen(ILI9341_BLACK);

  // draw lines
  tft.fillRect(0, 20, 320, 1, FOREGND_COLOUR);
  tft.fillRect(0, 120, 320, 1, FOREGND_COLOUR);
  tft.fillRect(0, 220, 320, 1, FOREGND_COLOUR);
  tft.fillRect(80, 20, 1, 200, FOREGND_COLOUR);
  tft.fillRect(160, 20, 1, 200, FOREGND_COLOUR);
  tft.fillRect(240, 20, 1, 200, FOREGND_COLOUR);

  drawVolumeBorders();
  drawButtons();
  
}

void loop() {

  readSerial();

  //update the LCD display at the set frame rate
  if ((millis() - previousMillis) > (1000.0 / (float)LCD_FRAME_RATE))
  {
    updateDisplay();
    previousMillis = millis();
  }
  
}

void updateDisplay() {

  drawPatchName();
  drawPatchSelected();
  drawKey();
  drawTrackNames();
  drawVolumes();
  drawButtons();
  
}

void drawPatchName() {

  if (current_patch_name != prev_current_patch_name) {
    prev_current_patch_name = current_patch_name;
    tft.fillRect(0, 0, 240, 20, ILI9341_BLACK);
    tft.fillRect(0, 221, 320, 19, ILI9341_BLACK); // also wipe over next patch text
    tft.setTextColor(ACTIVE_COLOUR, ILI9341_BLACK);
    tft.setCursor(10, 2); 
    tft.setTextSize(2);
    tft.print(current_patch_name);
  }
  
}

void drawPatchSelected() {

  if (next_patch_name != prev_next_patch_name) {
    prev_next_patch_name = next_patch_name;
    tft.fillRect(0, 221, 320, 19, ILI9341_BLACK);
    tft.setTextColor(FOREGND_COLOUR, ILI9341_BLACK);
    tft.setCursor(10, 222); 
    tft.setTextSize(2);
    tft.print(next_patch_name);
  }
  
}

void changeKey(String patchKey, int newKey) {

  int keyIndex = 0;
  for(int i = 0; i<12; i++) {
    if (keys[i] == patchKey) {
      keyIndex = i;
    }
  }

  int indexKey = keyIndex + newKey;
  if (indexKey < 0) {
    indexKey += 12;
  } else if (indexKey > 11) {
    indexKey -= 12;
  }

  if (indexKey != current_key) {
    current_key = indexKey;
  }
  
}

void drawKey() {

  if (current_key != prev_current_key) {
    prev_current_key = current_key;
    tft.fillRect(240, 0, 80, 20, ILI9341_BLACK);
    tft.setTextColor(ACTIVE_COLOUR, ILI9341_BLACK);
    tft.setCursor(240, 2); 
    tft.setTextSize(2);
    tft.print("key:" + keys[current_key]);
  }
  
}

void drawVolumeBorders() {

  for (uint8_t i = 0; i < NO_OF_TRACKS; i++) {
    int volY = 50;
    int volX;
    if (i >= 4) {
      volY = 150;
    }
    if (i < 4) {
      volX = 40 + i * 80;
    } else {
      volX = 40 + (i - 4) * 80;
    }
    tft.fillRect (volX - 30, volY, 60, 1, FOREGND_COLOUR);
    tft.fillRect (volX - 30, volY + 9, 60, 1, FOREGND_COLOUR);
    tft.fillRect (volX - 30, volY, 1, 10, FOREGND_COLOUR);
    tft.fillRect (volX - 30 + 60, volY, 1, 10, FOREGND_COLOUR);
  }
  
}

void drawVolumes() {

  for (uint8_t i = 0; i < NO_OF_TRACKS; i++) {
    drawVolume(i);
  }
  
}

void drawVolume(int index) {

  bool updateMinMarker = false;
  bool updateMaxMarker = false;
  bool updateVolume = false;
  int volY;
  int volX;

  if (track_auto_on[index] == 1) {
    if (track_min_volumes[index] != prev_track_min_volumes[index]) {
      updateMinMarker = true;
    }
    if (track_volumes[index] != prev_track_volumes[index]) {
      updateMaxMarker = true;
    }
  }
  if (track_volumes[index] != prev_track_volumes[index] || track_auto_volumes[index] != prev_track_auto_volumes[index]) {
    updateVolume = true;
  }

  if (updateMinMarker || updateMaxMarker || updateVolume) {
    
    volY = 50;
    if (index >= 4) {
      volY = 150;
    }
    if (index < 4) {
      volX = 40 + index * 80;
    } else {
      volX = 40 + (index - 4) * 80;
    }

    tft.fillRect (volX - 32, volY - 5, 64, 4, ILI9341_BLACK); // clear max and min volume marker area
    
  }

  // draw min and max volume markers when in automation mode
  if (updateMinMarker || updateMaxMarker) {
    tft.fillRect (volX - 30 + floor(track_min_volumes[index] * 60) - 2, volY - 5, 4, 4, FOREGND_COLOUR); // draw volume min marker
    tft.fillRect (volX - 30 + floor(track_volumes[index] * 60) - 2, volY - 5, 4, 4, FOREGND_COLOUR); // draw volume max marker
    if (updateMinMarker) {
      prev_track_min_volumes[index] = track_min_volumes[index];
    }
  }

  // draw track volume bar based on slider or actual volume if automated. also draw max volume marker when in automation mode (set by slider)
  if (updateVolume) {

    int currentVol;
    int previousVol;

    // don't completely redraw the bars each time
    // draw bars based on whether auto mode is on or not
    if (track_auto_on[index] == 1) {
      currentVol = floor(track_auto_volumes[index] * 60);
      previousVol = floor(prev_track_auto_volumes[index] * 60);
      if (currentVol > previousVol) {
        tft.fillRect (volX - 30 + 1, volY + 1, currentVol, 8, FOREGND_COLOUR);
      } else {
        tft.fillRect (volX - 30 + currentVol, volY + 1, (60 - currentVol), 8, ILI9341_BLACK);
      }
      prev_track_auto_volumes[index] = track_auto_volumes[index];
      //prev_track_volumes[index] = track_auto_volumes[index];
    } else {
      currentVol = floor(track_volumes[index] * 59);
      previousVol = floor(prev_track_volumes[index] * 59);
      if (currentVol >= previousVol) {
        tft.fillRect (volX - 29, volY + 1, currentVol, 8, FOREGND_COLOUR);
      } else {
        tft.fillRect (volX - 29 + currentVol, volY + 1, (59 - currentVol), 8, ILI9341_BLACK);
      }
      prev_track_volumes[index] = track_volumes[index];
      //prev_track_auto_volumes[index] = track_volumes[index];
    }

  }
  
}

void drawTrackNames() {

  for (uint8_t i = 0; i < NO_OF_TRACKS; i++) {
    drawTrackName(i);
  }
  
}

void drawTrackName(int index) {

  if (track_names[index] != prev_track_names[index]) {

    prev_track_names[index] = track_names[index];

    int nameX;
    int nameY = 30;
    if (index >= 4) {
      nameY = 130;
    }
    if (index < 4) {
      nameX = 10 + index * 80;
    } else {
      nameX = 10 + (index - 4) * 80;
    }
    tft.fillRect (nameX, nameY, 60, 8, ILI9341_BLACK);
    tft.setTextColor(ACTIVE_COLOUR, ILI9341_BLACK);
    tft.setCursor(nameX, nameY); 
    tft.setTextSize(1);
    tft.print(track_names[index]);

  }
  
}

void drawButtons() {

  for (uint8_t i = 0; i < NO_OF_TRACKS; i++) {
    drawAutoBtn(i);
    drawAutoAmount(i);
    drawFilterBtn(i);
    drawFilterAmount(i);
    drawDistortionBtn(i);
    drawDistortionAmount(i);
    drawTrackMode(i);
  }
  
}

void drawTrackMode(int index) {

  if (track_modes[index] != prev_track_modes[index]) {

    prev_track_modes[index] = track_modes[index];

    int modeX;
    int modeY = 72;
    if (index >= 4) {
      modeY = 172;
    }
    if (index < 4) {
      modeX = 5 + index * 80;
    } else {
      modeX = 5 + (index - 4) * 80;
    }
  
    tft.fillRect(modeX - 2, modeY - 22, 6, 68, ILI9341_BLACK);
  
    modeY += (track_modes[index] * 18);
    if (track_modes[index] == 3) {
      modeY -= (4 * 18);
    }
    tft.fillCircle(modeX, modeY, 2.5, FOREGND_COLOUR); // x, y, radius, color

  }
  
}

void drawAutoBtn(int index) {

  if (track_auto_on[index] != prev_track_auto_on[index]) {

    prev_track_auto_on[index] = track_auto_on[index];

    int autoX;
    int autoY = 70;
    if (index >= 4) {
      autoY = 170;
    }
    if (index < 4) {
      autoX = 12 + index * 80;
    } else {
      autoX = 12 + (index - 4) * 80;
    }
  
    if (track_auto_on[index] == 1) {
      tft.fillRect(autoX - 2, autoY - 2, 10, 12, FOREGND_COLOUR);
      tft.setTextColor(ILI9341_BLACK, FOREGND_COLOUR);
    } else {
      tft.fillRect(autoX - 2, autoY - 2, 10, 12, ILI9341_BLACK);
      tft.setTextColor(FOREGND_COLOUR, ILI9341_BLACK);
    }
    tft.setCursor(autoX, autoY); 
    tft.setTextSize(1);
    tft.print("A");

  }
  
}

void drawAutoAmounts() {

  for (uint8_t i = 0; i < NO_OF_TRACKS; i++) {
    drawAutoAmount(i);
  }
  
}

void drawAutoAmount(int index) {

  if (track_auto_amount[index] != prev_track_auto_amount[index]) {

    prev_track_auto_amount[index] = track_auto_amount[index];

    int barY = 70;
    int barX;
    if (index >= 4) {
      barY = 170;
    }
    if (index < 4) {
      barX = 25 + index * 80;
    } else {
      barX = 25 + (index - 4) * 80;
    }
    int barWidth = 45 - floor(track_auto_amount[index] / 2000 * 45); // inverted bar - higher value is slower
    tft.fillRect (barX, barY, 45, 4, ILI9341_BLACK); 
    tft.fillRect (barX, barY, barWidth, 4, FOREGND_COLOUR); 

  }
  
}

void drawFilterBtn(int index) {

  if (track_filter_on[index] != prev_track_filter_on[index]) {

    prev_track_filter_on[index] = track_filter_on[index];

    int autoX;
    int autoY = 88;
    if (index >= 4) {
      autoY = 188;
    }
    if (index < 4) {
      autoX = 12 + index * 80;
    } else {
      autoX = 12 + (index - 4) * 80;
    }
    if (track_filter_on[index] == 1) {
      tft.fillRect(autoX - 2, autoY - 2, 10, 12, FOREGND_COLOUR);
      tft.setTextColor(ILI9341_BLACK, FOREGND_COLOUR);
    } else {
      tft.fillRect(autoX - 2, autoY - 2, 10, 12, ILI9341_BLACK);
      tft.setTextColor(FOREGND_COLOUR, ILI9341_BLACK);
    }
    tft.setCursor(autoX, autoY); 
    tft.setTextSize(1);
    tft.print("F");

  }
  
}

void drawFilterAmounts() {

  for (uint8_t i = 0; i < NO_OF_TRACKS; i++) {
    drawFilterAmount(i);
  }
  
}

void drawFilterAmount(int index) {

  if (track_filter_amount[index] != prev_track_filter_amount[index]) {

    prev_track_filter_amount[index] = track_filter_amount[index];

    int barY = 88;
    int barX;
    if (index >= 4) {
      barY = 188;
    }
    if (index < 4) {
      barX = 25 + index * 80;
    } else {
      barX = 25 + (index - 4) * 80;
    }
    int barWidth = floor(track_filter_amount[index] / 12500 * 45);
    tft.fillRect (barX, barY, 45, 4, ILI9341_BLACK); 
    tft.fillRect (barX, barY, barWidth, 4, FOREGND_COLOUR); 

  }
  
}

void drawDistortionBtn(int index) {

  if (track_distortion_on[index] != prev_track_distortion_on[index]) {

    prev_track_distortion_on[index] = track_distortion_on[index];

    int autoX;
    int autoY = 106;
    if (index >= 4) {
      autoY = 206;
    }
    if (index < 4) {
      autoX = 12 + index * 80;
    } else {
      autoX = 12 + (index - 4) * 80;
    }
    if (track_distortion_on[index] == 1) {
      tft.fillRect(autoX - 2, autoY - 2, 10, 12, FOREGND_COLOUR);
      tft.setTextColor(ILI9341_BLACK, FOREGND_COLOUR);
    } else {
      tft.fillRect(autoX - 2, autoY - 2, 10, 12, ILI9341_BLACK);
      tft.setTextColor(FOREGND_COLOUR, ILI9341_BLACK);
    }
    tft.setCursor(autoX, autoY); 
    tft.setTextSize(1);
    tft.print("D");

  }
  
}

void drawDistortionAmounts() {

  for (uint8_t i = 0; i < NO_OF_TRACKS; i++) {
    drawDistortionAmount(i);
  }
  
}

void drawDistortionAmount(int index) {

  if (track_distortion_amount[index] != prev_track_distortion_amount[index]) {

    prev_track_distortion_amount[index] = track_distortion_amount[index];

    int barY = 106;
    int barX;
    if (index >= 4) {
      barY = 206;
    }
    if (index < 4) {
      barX = 25 + index * 80;
    } else {
      barX = 25 + (index - 4) * 80;
    }
    int barWidth = floor(track_distortion_amount[index] / 127 * 45);
    tft.fillRect (barX, barY, 45, 4, ILI9341_BLACK); 
    tft.fillRect (barX, barY, barWidth, 4, FOREGND_COLOUR); 

  }
  
}

void readSerial() {

  //Read Serial
  while(Serial.available() > 0)
  {
    char c = Serial.read();
    if(c=='#'){
      serIndex = 0;
      inData[serIndex] = '\0';
      serialStarted = true;
    }else if(c=='/'){
      serialEnded = true;
      break;
    }else{
      if(serIndex < 79){
        inData[serIndex] = c;
        serIndex++;
        inData[serIndex] = '\0';
      }
    }
  }

  //Process Serial Package
  if(serialStarted && serialEnded){

    String PrintString = String(inData);

    int commaIndex = PrintString.indexOf('_');
    int secondCommaIndex = PrintString.indexOf('_', commaIndex+1);
    int thirdCommaIndex = PrintString.indexOf('_', secondCommaIndex+1);

    String part1 = PrintString.substring(0, commaIndex);
    String part2 = PrintString.substring(commaIndex+1, secondCommaIndex);
    String part3 = PrintString.substring(secondCommaIndex+1, thirdCommaIndex);

    //testTEXT
    //debugMessage(part1, part2, part3);

    // track volume (or max volume if automated)
    if (part1.equals("vol")) {
      int trackVol = part2.toFloat();
      float valueVol = part3.toFloat();
      if (track_volumes[trackVol] != valueVol) {
        track_volumes[trackVol] = valueVol;
      }
    }

    // auto volume (or actual volume if automated)
    if (part1.equals("voo")) {
      int trackAutoVol = part2.toFloat();
      float valueAutoVol = part3.toFloat();
      if (track_auto_volumes[trackAutoVol] != valueAutoVol) {
        track_auto_volumes[trackAutoVol] = valueAutoVol;
      }
    }

    // min volume (used in automation)
    if (part1.equals("vmn")) {
      int trackMinVol = part2.toFloat();
      float valueMinVol = part3.toFloat();
      if (track_min_volumes[trackMinVol] != valueMinVol) {
        track_min_volumes[trackMinVol] = valueMinVol;
      }
    }

    // track mode (0 / 1 / 2 / 3)
    if (part1.equals("mod")) {
      int trackMode = part2.toFloat();
      int trackModeValue = part3.toFloat();
      if (track_modes[trackMode] != trackModeValue) {
        track_modes[trackMode] = trackModeValue;
      }
    }

    // track auto toggle
    if (part1.equals("aon")) {
      int trackAutoOn = part2.toFloat();
      int valueAutoOn = part3.toFloat();
      if (track_auto_on[trackAutoOn] != valueAutoOn) {
        track_auto_on[trackAutoOn] = valueAutoOn;
      }
    }

    // track auto amount
    if (part1.equals("aam")) {
      int trackAutoAm = part2.toFloat();
      int valueAutoAm = part3.toFloat();
      if (track_auto_amount[trackAutoAm] != valueAutoAm) {
        track_auto_amount[trackAutoAm] = valueAutoAm;
      }
    }

    // track filter toggle
    if (part1.equals("fon")) {
      int trackFilterOn = part2.toFloat();
      int valueFilterOn = part3.toFloat();
      if (track_filter_on[trackFilterOn] != valueFilterOn) {
        track_filter_on[trackFilterOn] = valueFilterOn;
      }
    }

    // track filter amount
    if (part1.equals("fam")) {
      int trackFilterAm = part2.toFloat();
      int valueFilterAm = part3.toFloat();
      if (track_filter_amount[trackFilterAm] != valueFilterAm) {
        track_filter_amount[trackFilterAm] = valueFilterAm;
      }
    }

    // track distortion toggle
    if (part1.equals("don")) {
      int trackDistortionOn = part2.toFloat();
      int valueDistortionOn = part3.toFloat();
      if (track_distortion_on[trackDistortionOn] != valueDistortionOn) {
        track_distortion_on[trackDistortionOn] = valueDistortionOn;
      }
    }

    // track distortion amount
    if (part1.equals("dam")) {
      int trackDistortionAm = part2.toFloat();
      int valueDistortionAm = part3.toFloat();
      if (track_distortion_amount[trackDistortionAm] != valueDistortionAm) {
        track_distortion_amount[trackDistortionAm] = valueDistortionAm;
      }
    }

    // load new patch
    if (part1.equals("new")) {
      if (part3 != current_patch_name) {
        current_patch_name = part2 + ":" + part3;
      }
    }

    // pick next/prev patch
    if (part1.equals("nxt")) {
      if (part3 != next_patch_name) {
        next_patch_name = part2 + ":" + part3;
      }
    }

    // track name
    if (part1.equals("nam")) {
      int trackNameIndex = part2.toFloat();
      if (part3 != track_names[trackNameIndex]) {
        track_names[trackNameIndex] = part3;
      }
    }

    // transpose key
    if (part1.equals("key")) {
      int newKey = part3.toFloat();
      //if (newKey != current_key) {
        changeKey(part2, newKey);
      //}
    }
    
  }
}

void debugMessage(String pt1, String pt2, String pt3){
  tft.fillRect(0, 100, 320, 240, 0x033f);
  tft.setCursor(40, 120);
  tft.setTextColor(ILI9341_WHITE);
  tft.print(pt1);
  tft.setCursor(40, 140);
  tft.print(pt2);
  tft.setCursor(40, 160);
  tft.print(pt3);  
}
